/* 
 * Copyright (c) Holger Pfaff - http://pfaff.ws
 *
 * This software maybe used for any purpose provided the
 * above copyright notice is retained. It is supplied as is.
 * No warranty expressed or implied - Use at your own risk.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>

/* Progname */
#define PROGNAME "x-fontperf"
#define PROGVERSION VERSION
#define PROGDATE VERSIONDATE

/* size of string buffer */
#define MAXBUF 256

/* max number of returns for XListFonts */
#define MAXFONTS 100000

/* default X DISPLAY to use */
#define DISPLAY      ""

/* defaults for Tests 1 & 2 */
#define FOUNDRY      "*"
#define FAMILY       "helvetica"
#define WEIGHT       "*"
#define SLANT        "*"
#define SETWIDTH     "*"
#define ADDSTYLE     "*"
#define PIXELSIZE    "12"
#define POINTSIZE    "*"
#define RESOLUTIONX  "*"
#define RESOLUTIONY  "*"
#define SPACING      "*"
#define AVGWIDTH     "*"
#define REGISTRY     "iso8859"
#define ENCODING     "1"

/* defaults for Test 1 */
#define ITERATIONS   100

/* defaults for Test 2 */
#define MINSIZE  6
#define MAXSIZE  144

#define GETTIME(t) gettimeofday(&t, NULL)
#define DIFFTIME(t1, t2) (t2.tv_sec - t1.tv_sec) * 1000000 + (t2.tv_usec - t1.tv_usec);

void usage(char *progname) {

  fprintf(stderr, "\nUsage: %s ...\n\n", progname);

  fprintf(stderr, "  -display <dsp>   X-display to use\n\n");

  fprintf(stderr, "  -v               Increase verbosity (may falsify results)\n");
  fprintf(stderr, "  -e               Show errors during tests\n\n");

  fprintf(stderr, "  -n <iter>        Number of iterations for test 1 [%d]\n\n", ITERATIONS);

  fprintf(stderr, "  -min <pixel>     Minimum pixel size for test 2 [%d]\n", MINSIZE);
  fprintf(stderr, "  -max <pixel>     Maximum pixel size for test 2 [%d]\n\n", MAXSIZE);

  fprintf(stderr, "  -foundry <fndr>  Foundry of font [%s]\n", FOUNDRY);
  fprintf(stderr, "  -family <fmly>   Family of font [%s]\n", FAMILY);
  fprintf(stderr, "  -weight <wght>   Weight of font [%s]\n", WEIGHT);
  fprintf(stderr, "  -slant <slnt>    Slant of font [%s]\n", SLANT);
  fprintf(stderr, "  -setwidth <wdth> Set width of font [%s]\n", SETWIDTH);
  fprintf(stderr, "  -addstyle <styl> Additional style of font [%s]\n", ADDSTYLE);
  fprintf(stderr, "  -pixelsize <sze> Pixel size of font [%s]\n", PIXELSIZE);
  fprintf(stderr, "  -pointsize <sze> Point size of font [%s]\n", POINTSIZE);
  fprintf(stderr, "  -resolutionX <r> X resolution of font [%s]\n", RESOLUTIONX);
  fprintf(stderr, "  -resolutionY <r> Y resolution of font [%s]\n", RESOLUTIONY);
  fprintf(stderr, "  -spacing <spc>   Spacing of font [%s]\n", SPACING);
  fprintf(stderr, "  -avgwidth <aw>   Average width of font [%s]\n", AVGWIDTH);
  fprintf(stderr, "  -registry <reg>  Registry of font [%s]\n", REGISTRY);
  fprintf(stderr, "  -encoding <enc>  Encoding of font [%s]\n\n", ENCODING);

  exit(1);
}

int main(int argc, char *argv[]) {

  char  *progname    = PROGNAME;
  char  *progversion = PROGVERSION;
  char  *progdate    = PROGDATE;
  
  int    i;
  int    usec;

  /* returns from XListFonts */
  int    nfonts;
  char **fonts;

  /* returns from XGetFontPath */
  int    nfontpaths;
  char **fontpaths;
  
  /* font spec */
  char  *foundry = FOUNDRY;
  char  *family = FAMILY;
  char  *weight = WEIGHT;
  char  *slant = SLANT;
  char  *setwidth = SETWIDTH;
  char  *addstyle = ADDSTYLE;
  char  *pixelsize = PIXELSIZE;
  char  *pointsize = POINTSIZE;
  char  *resolutionX = RESOLUTIONX;
  char  *resolutionY = RESOLUTIONY;
  char  *spacing = SPACING;
  char  *avgwidth = AVGWIDTH;
  char  *registry = REGISTRY;
  char  *encoding = ENCODING;
  
  /* values for Test 1 */
  int    iterations = ITERATIONS;
  
  /* values for Test 2 */
  int    minsize = MINSIZE;
  int    maxsize = MAXSIZE;
  
  /* display to use */
  char  *display = DISPLAY;
  
  /* char buffer for font names */
  char  buffer[MAXBUF];
  
  /* timestamps */
  struct timeval t1, t2, t3;
  
  /* verbosity increases with value */
  int verbose = 0;

  /* show errors during test ? */
  int showerror = 0;

  /* return from XLoadQueryFont */
  XFontStruct *fs;
  
  /* the display handle */
  Display *Display;
  
  printf("%s: Copyright (c) Holger Pfaff - http://pfaff.ws\n", progname);
  
  for(i = 1; i < argc; ++i) {
    if(strcmp(argv[i], "-v") == 0) {
      ++verbose;
    } else if(strcmp(argv[i], "-e") == 0) {
      showerror = 1;
    } else if(strcmp(argv[i], "-n") == 0 && i + 1 < argc) {
      iterations = atoi(argv[++i]);
    } else if(strcmp(argv[i], "-min") == 0 && i + 1 < argc) {
      minsize = atoi(argv[++i]);
    } else if(strcmp(argv[i], "-max") == 0 && i + 1 < argc) {
      maxsize = atoi(argv[++i]);
    } else if(strcmp(argv[i], "-display") == 0 && i + 1 < argc) {
      display = argv[++i];
    } else if(strcmp(argv[i], "-foundry") == 0 && i + 1 < argc) {
      foundry = argv[++i];
    } else if(strcmp(argv[i], "-family") == 0 && i + 1 < argc) {
      family = argv[++i];
    } else if(strcmp(argv[i], "-weight") == 0 && i + 1 < argc) {
      weight = argv[++i];
    } else if(strcmp(argv[i], "-slant") == 0 && i + 1 < argc) {
      slant = argv[++i];
    } else if(strcmp(argv[i], "-setwidth") == 0 && i + 1 < argc) {
      setwidth = argv[++i];
    } else if(strcmp(argv[i], "-addstyle") == 0 && i + 1 < argc) {
      addstyle = argv[++i];
    } else if(strcmp(argv[i], "-pixelsize") == 0 && i + 1 < argc) {
      pixelsize = argv[++i];
    } else if(strcmp(argv[i], "-pointsize") == 0 && i + 1 < argc) {
      pointsize = argv[++i];
    } else if(strcmp(argv[i], "-resolutionX") == 0 && i + 1 < argc) {
      resolutionX = argv[++i];
    } else if(strcmp(argv[i], "-resolutionY") == 0 && i + 1 < argc) {
      resolutionY = argv[++i];
    } else if(strcmp(argv[i], "-spacing") == 0 && i + 1 < argc) {
      spacing = argv[++i];
    } else if(strcmp(argv[i], "-avgwidth") == 0 && i + 1 < argc) {
      avgwidth = argv[++i];
    } else if(strcmp(argv[i], "-registry") == 0 && i + 1 < argc) {
      registry = argv[++i];
    } else if(strcmp(argv[i], "-encoding") == 0 && i + 1 < argc) {
      encoding = argv[++i];
    } else {
      usage(progname);
    }
  }

  if(verbose > 0) {
    printf("%s: version %s from %s\n", progname, progversion, progdate);
  }

  /* open the Display */
  Display = XOpenDisplay(display);
  if(Display == 0) {
    fprintf(stderr, "%s: unable to open display '%s'\n", progname, XDisplayName(display));
    exit(1);
  }

  if(verbose > 0) {
    printf("%s: connected to server: %s\n", progname, DisplayString(Display));
    printf("%s: server vendor: %s\n", progname, ServerVendor(Display));
    printf("%s: vendor release: %d\n", progname, VendorRelease(Display));
  }

  /* get complete fontlist */                        
  fonts = XListFonts(Display, "*", MAXFONTS, &nfonts);
  if(verbose > 0) {
    printf("%s: total of %d fonts found\n", progname, nfonts);
  }

  /* obtain fontpaths */
  if(verbose > 1) {
    fontpaths = XGetFontPath(Display, &nfontpaths);
    for(i = 0; i < nfontpaths; ++i) {
      printf("%s: fontpath %d: %s\n", progname, i, fontpaths[i]);
    }
  }

  /* Test 1: loading same font many times */
  sprintf(buffer, "-%s-%s-%s-%s-%s-%s-%s-%s-%s-%s-%s-%s-%s-%s", 
                   foundry, family, weight, slant, setwidth, addstyle,
                   pixelsize, pointsize, resolutionX, resolutionY,
                   spacing, avgwidth, registry, encoding);
  printf("%s: basefont: %s\n", progname, buffer);
  printf("%s: loading basefont %d times ... ", progname, iterations); fflush(stdout);
  GETTIME(t1); 
  for(i = 0; i < iterations; ++i) {
    fs = XLoadQueryFont(Display, buffer);
    if(fs) { 
      XFreeFont(Display, fs);
    } else {
      if(showerror) {
        fprintf(stderr, "%s: error open font: %s\n", progname, buffer);
      }
    }
  }
  GETTIME(t2); 
  usec = DIFFTIME(t1, t2);
  printf("%d usec %d usec/font\n", usec, usec/iterations);
  
  /* Test 2: loading same font from small to large */
  printf("%s: loading font from pixel size %d to %d ... ", progname, minsize, maxsize); fflush(stdout);
  GETTIME(t1); 
  GETTIME(t3); 
  for(i = minsize; i <= maxsize; ++i) {
    /* sprintf only takes about 6 usec on my 400MHz Sun */
    sprintf(buffer, "-%s-%s-%s-%s-%s-%s-%d-%s-%s-%s-%s-%s-%s-%s", 
                     foundry, family, weight, slant, setwidth, addstyle,
                     i, pointsize, resolutionX, resolutionY,
                     spacing, avgwidth, registry, encoding);
    if(verbose > 2) {
      printf("%s: loading font %s ... ", progname, buffer); fflush(stdout);
    }
    fs = XLoadQueryFont(Display, buffer);
    if(fs) {
      XFreeFont(Display, fs);
    } else {
      if(showerror) {
        fprintf(stderr, "%s: error open font: %s\n", progname, buffer);
      }
    }
    if(verbose > 2) {
      GETTIME(t2); 
      usec = DIFFTIME(t3, t2);
      printf("%d usec\n", usec);
      GETTIME(t3); 
    }
  }
  GETTIME(t2); 
  usec = DIFFTIME(t1, t2);
  printf("%d usec %d usec/font\n", usec, usec/(maxsize - minsize));

  /* Test 3: loading all available fonts */
  printf("%s: loading all %d available fonts ... ", progname, nfonts); fflush(stdout);
  GETTIME(t1); 
  GETTIME(t3); 
  for(i = 0; i < nfonts; ++i) {
    if(verbose > 2) {
      printf("%s: loading font %s ... ", progname, fonts[i]); fflush(stdout);
    }
    fs = XLoadQueryFont(Display, fonts[i]);
    if(fs) {
      XFreeFont(Display, fs);
    } else {
      if(showerror) {
        fprintf(stderr, "%s: error open font: %s %d\n", progname, fonts[i], i);
      }
    }
    if(verbose > 2) {
      GETTIME(t2); 
      usec = DIFFTIME(t3, t2);
      printf("%d usec\n", usec);
      GETTIME(t3); 
    }
  }
  GETTIME(t2); 
  usec = DIFFTIME(t1, t2);
  printf("%d usec %d usec/font\n", usec, usec/nfonts);

  exit(0);
}
